// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2009-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using Mono.Unix;

namespace Smdn.IO.FileSystems.Unix {
  public class UnixDirectory : Directory, IUnixFileSystemEntry {
    public override string Name {
      get { return info.Name; }
    }

    public override long Length {
      get {
        try {
          return info.Length;
        }
        catch (InvalidOperationException) {
          return 0L;
        }
      }
    }

    public override bool Exists {
      get { return info.Exists; }
    }

    public override bool IsRoot {
      get { return info.Parent == null; }
    }

    public override Directory Parent {
      get { return (info.Parent == null) ? null : new UnixDirectory(info.Parent); }
    }

    public override Directory Root {
      get { return new UnixDirectory(info.Root); }
    }

    public override string FullName {
      get { return info.FullName; }
    }

    public override object RawInfo {
      get { return info; }
    }

    UnixFileSystemInfo IUnixFileSystemEntry.Info {
      get { return info; }
    }

    internal UnixDirectory(UnixDirectoryInfo directoryInfo)
    {
      if (directoryInfo == null)
        throw new ArgumentNullException("directoryInfo");

      this.info = directoryInfo;
    }

    public override long GetLength()
    {
      return info.Length;
    }

    public override void Refresh()
    {
      info.Refresh();
    }

    private IEnumerable<UnixFileSystemInfo> GetFileSystemEntries(ScanDirectoryExceptionHandler exceptionHandler)
    {
      try {
        return info.GetFileSystemEntries();
      }
      catch (Exception ex) {
        if (exceptionHandler == null)
          throw ex;
        else
          exceptionHandler(this, ex);

        return EmptyArray<UnixFileSystemInfo>.Instance;
      }
    }

    public override IEnumerable<DirectoryEntry> GetEntries(ScanDirectoryExceptionHandler exceptionHandler, Predicate<DirectoryEntry> match)
    {
      foreach (var entry in GetFileSystemEntries(exceptionHandler)) {
        DirectoryEntry e = null;

        if (entry.IsRegularFile)
          e = new UnixFile(entry as UnixFileInfo);
        else if (entry.IsDirectory)
          e = new UnixDirectory(entry as UnixDirectoryInfo);
        else
          continue;

        if (match == null || match(e))
          yield return e;
      }
    }

    public override IEnumerable<File> GetFiles(ScanDirectoryExceptionHandler exceptionHandler, Predicate<File> match)
    {
      foreach (var entry in GetFileSystemEntries(exceptionHandler)) {
        if (!entry.IsRegularFile)
          continue;

        var f = new UnixFile(entry as UnixFileInfo);

        if (match == null || match(f))
          yield return f;
      }
    }

    public override IEnumerable<Directory> GetDirectories(ScanDirectoryExceptionHandler exceptionHandler, Predicate<Directory> match)
    {
      foreach (var entry in GetFileSystemEntries(exceptionHandler)) {
        if (!entry.IsDirectory)
          continue;

        var d = new UnixDirectory(entry as UnixDirectoryInfo);

        if (match == null || match(d))
          yield return d;
      }
    }

    private UnixDirectoryInfo info;
  }
}
