//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2003-2014 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Drawing;
using System.Reflection;

using Smdn.Collections;
using Smdn.IO.FileSystems;
using Smdn.Windows.UserInterfaces.Shells;

namespace Smdn.Applications.BittenBaum.UI {
  using Ico = Smdn.Imaging.Formats.Ico;

  public static class IconTheme {
    private interface ISystemIconTheme {
      Bitmap GetIcon(IconType type);
      Bitmap GetIcon(Directory directory);
      Bitmap GetIcon(Drive drive);
    }

    private class WindowsIconTheme : ISystemIconTheme {
      private static readonly string pathToShell32dll = Environment.ExpandEnvironmentVariables("%systemroot%\\system32\\shell32.dll");
      private static readonly string pathToMydocsdll = Environment.ExpandEnvironmentVariables("%systemroot%\\system32\\mydocs.dll");
      private static readonly string pathToExplorer = Environment.ExpandEnvironmentVariables("%systemroot%\\explorer.exe");

      public WindowsIconTheme()
      {
      }

      Bitmap ISystemIconTheme.GetIcon(IconType type)
      {
        switch (type) {
          case IconType.Folder:             return GetIconFromFile(pathToExplorer, 13);
          case IconType.DesktopFolder:      return GetIconFromFile(pathToExplorer, 3);
          case IconType.MyDocumentFolder:   return GetIconFromFile(pathToMydocsdll, 0);
          case IconType.ProgramFilesFolder: return GetIconFromFile(pathToShell32dll, 19);
          case IconType.FolderSelection:    return GetIconFromFile(pathToShell32dll, 22);
          default: return null;
        }
      }

      Bitmap ISystemIconTheme.GetIcon(Directory directory)
      {
        try {
          var desktopIni = new DesktopIni(directory.FullName);

          return
            desktopIni.IconResource.Extract() ??
            desktopIni.IconLocation.Extract() ??
            GetIcon(IconType.Folder);
        }
        catch {
          return GetIcon(IconType.Folder);
        }
      }

      Bitmap ISystemIconTheme.GetIcon(Drive drive)
      {
        switch (drive.Type) {
          case System.IO.DriveType.Fixed:       return GetIconFromFile(pathToShell32dll, 8);
          case System.IO.DriveType.CDRom:       return GetIconFromFile(pathToShell32dll, 11);
          case System.IO.DriveType.Network:     return GetIconFromFile(pathToShell32dll, 9);
          case System.IO.DriveType.Removable:   return GetIconFromFile(pathToShell32dll, 7);
          default: return GetIcon(IconType.Folder); // フォルダのアイコンで代用
        }
      }

      private static Bitmap GetIconFromFile(string file, int index)
      {
        try {
          return Ico.Icon.Extract(file, index, 1, Ico.IconSize.Small, true).First();
        }
        catch {
          return null;
        }
      }
    }

    private class GnomeIconTheme : ISystemIconTheme {
      private static readonly MethodInfo getIconLocation;

      static GnomeIconTheme()
      {
        getIconLocation = null;

        if (!Runtime.IsRunningOnMono)
          return;

        try {
          const string code = @"
namespace Smdn.Applications.BittenBaum.Interop {
  public static class IconTheme {
    public static string GetIconLocation(string iconName)
    {
      Gtk.IconInfo iconInfo = Gtk.IconTheme.Default.LookupIcon(iconName, 16, Gtk.IconLookupFlags.NoSvg);

      if (iconInfo == null)
        return null;
      else
        return iconInfo.Filename;
    }
  }
}";

          using (var provider = new Microsoft.CSharp.CSharpCodeProvider(new System.Collections.Generic.Dictionary<string, string>() {{"CompilerVersion", "v2.0"}})) {
            var options = new System.CodeDom.Compiler.CompilerParameters();

            options.CompilerOptions = "-pkg:gtk-sharp-2.0";
            options.GenerateInMemory = true;
            options.IncludeDebugInformation = false;

            var result = provider.CompileAssemblyFromSource(options, code);

            if (result.CompiledAssembly == null) {
              foreach (var error in result.Errors) {
                Console.Error.WriteLine(error);
              }
            }
            else {
              var type = result.CompiledAssembly.GetType("Smdn.Applications.BittenBaum.Interop.IconTheme", false);

              if (type != null)
                getIconLocation = type.GetMethod("GetIconLocation", BindingFlags.Static | BindingFlags.Public);
            }
          }
        }
        catch (Exception ex) {
          // ignore exceptions
          Console.Error.WriteLine(ex);
        }
      }

      private static Bitmap LookupIcon(string iconName)
      {
        try {
          var iconLocation = GetIconFileName(iconName);

          if (iconLocation == null)
            return null;

          using (var stream = System.IO.File.OpenRead(iconLocation)) {
            return new Bitmap(stream);
          }
        }
        catch (Exception ex) {
          Console.Error.WriteLine("Couldn't load icon {0}: {1}", iconName, ex);
          return null;
        }
      }

      private static string GetIconFileName(string iconName)
      {
        try {
          if (getIconLocation == null)
            return null;
          else
            return getIconLocation.Invoke(null, new object[] {iconName}) as string;
        }
        catch (Exception ex) {
          Console.Error.WriteLine(ex);
          return null;
        }
      }

      public GnomeIconTheme()
      {
        if (getIconLocation == null)
          throw new PlatformNotSupportedException();
      }

      Bitmap ISystemIconTheme.GetIcon(IconType type)
      {
        switch (type) {
          case IconType.Folder:           return LookupIcon("folder");
          case IconType.DesktopFolder:    return LookupIcon("desktop");
          case IconType.MyDocumentFolder: return LookupIcon("folder_home");
          case IconType.FolderSelection:  return LookupIcon("fileopen");
          default: return null;
        }
      }

      Bitmap ISystemIconTheme.GetIcon(Directory directory)
      {
        if (System.IO.Path.Equals(Environment.GetFolderPath(Environment.SpecialFolder.Desktop), directory.FullName))
          return GetIcon(IconType.DesktopFolder);
        else if (System.IO.Path.Equals(Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments), directory.FullName))
          return GetIcon(IconType.MyDocumentFolder);
        else
          return GetIcon(IconType.Folder);
      }

      Bitmap ISystemIconTheme.GetIcon(Drive drive)
      {
        // TODO
        return GetIcon(IconType.Folder);
      }
    }

    private class AlternativeIconTheme : ISystemIconTheme {
      private static readonly Bitmap Empty = new Bitmap(16, 16);

      private static Bitmap LoadIconFromResource(string iconName)
      {
        var resourceName = "Smdn.Applications.BittenBaum.res.icons." + iconName;

        using (var stream = Assembly.GetExecutingAssembly().GetManifestResourceStream(resourceName)) {
          if (stream == null)
            return Empty;

          try {
            return new Bitmap(stream);
          }
          catch {
            // ignore exceptions
            return Empty;
          }
        }
      }

      Bitmap ISystemIconTheme.GetIcon(IconType type)
      {
        switch (type) {
          case IconType.FolderSelection: return LoadIconFromResource("folder_explore.png");
          case IconType.MyDocumentFolder: return LoadIconFromResource("folder_user.png");
          case IconType.Folder:
          default:
            return LoadIconFromResource("folder.png");
        }
      }

      Bitmap ISystemIconTheme.GetIcon(Directory directory)
      {
        return GetIcon(IconType.Folder);
      }

      Bitmap ISystemIconTheme.GetIcon(Drive drive)
      {
        switch (drive.Type) {
          case System.IO.DriveType.Fixed:       return LoadIconFromResource("drive.png");
          case System.IO.DriveType.CDRom:       return LoadIconFromResource("drive_cd.png");
          case System.IO.DriveType.Network:     return LoadIconFromResource("drive_network.png");
          case System.IO.DriveType.Removable:   return LoadIconFromResource("drive_disk.png");
          default: return GetIcon(IconType.Folder); // フォルダのアイコンで代用
        }
      }
    }

    private static readonly ISystemIconTheme iconTheme;
    private static readonly ISystemIconTheme alternativeIconTheme = new AlternativeIconTheme();

    static IconTheme()
    {
      if (Runtime.IsRunningOnWindows) {
        iconTheme = new WindowsIconTheme();
      }
      else {
        try {
          // TODO: KDE, etc.
          iconTheme = new GnomeIconTheme();
        }
        catch (PlatformNotSupportedException) {
        }
        catch (Exception ex) {
          Console.Error.WriteLine(ex);
        }

        if (iconTheme == null)
          iconTheme = alternativeIconTheme;
      }
    }

    public static Bitmap GetIcon(IconType type)
    {
      return iconTheme.GetIcon(type) ?? alternativeIconTheme.GetIcon(type);
    }

    public static Bitmap GetIcon(Directory directory)
    {
      return iconTheme.GetIcon(directory) ?? alternativeIconTheme.GetIcon(directory);
    }

    public static Bitmap GetIcon(Drive drive)
    {
      return iconTheme.GetIcon(drive) ?? alternativeIconTheme.GetIcon(drive);
    }
  }
}
