//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2003-2014 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Drawing;
using System.Text.RegularExpressions;
using System.Windows.Forms;

using Smdn.IO.FileSystems;
using Smdn.Windows.Forms;

namespace Smdn.Applications.BittenBaum.UI {
  using Analysis;
  using Graphing;

  public partial class MainWindow {
    private GraphElementContextMenuStrip directoryContextMemuStrip = null;
    private GraphElementContextMenuStrip fileContextMemuStrip = null;
    private DirectoryTreeContextMenuStrip rootDirectoryContextMemuStrip = null;

    private void InitializeHandler()
    {
      // 保存
      toolStripMenuItemFileSave_Graph.Click += delegate { SaveResult(SaveResultFormatType.Graph); };
      toolStripMenuItemFileSave_Csv.Click += delegate { SaveResult(SaveResultFormatType.Csv); };
      toolStripMenuItemFileSave_Xml.Click += delegate { SaveResult(SaveResultFormatType.Xml); };
      
      // 終了
      toolStripMenuItemFile_Quit.Click += delegate { Close(); };

      // グラフに関するオプション
      toolStripMenuItemOption_Graph.Click += delegate { ConfigureGraphOptions(); };

      // 検索結果に関するオプション
      toolStripMenuItemOption_Tabulation.DropDownOpening += delegate {
        foreach (ToolStripItem item in toolStripMenuItemOption_Tabulation.DropDownItems) {
          if (item is ToolStripMenuItem)
            (item as ToolStripMenuItem).Checked = false;
        }

        switch (tabulator.TabulateMax) {
          case  10: toolStripMenuItemOptionTabulation_Upto10.Checked = true; break;
          case  20: toolStripMenuItemOptionTabulation_Upto20.Checked = true; break;
          case  30: toolStripMenuItemOptionTabulation_Upto30.Checked = true; break;
          case  50: toolStripMenuItemOptionTabulation_Upto50.Checked = true; break;
          case 100: toolStripMenuItemOptionTabulation_Upto100.Checked = true; break;
          case   0: toolStripMenuItemOptionTabulation_All.Checked = true; break;
          default:  toolStripMenuItemOptionTabulation_Upto30.Checked = true; break;
        }
      };

      foreach (var pair in new[] {
        new { MenuItem = toolStripMenuItemOptionTabulation_Upto10, Value = 10 },
        new { MenuItem = toolStripMenuItemOptionTabulation_Upto20, Value = 20 },
        new { MenuItem = toolStripMenuItemOptionTabulation_Upto30, Value = 30 },
        new { MenuItem = toolStripMenuItemOptionTabulation_Upto50, Value = 50 },
        new { MenuItem = toolStripMenuItemOptionTabulation_Upto100, Value = 100},
        new { MenuItem = toolStripMenuItemOptionTabulation_All, Value = 0 },
      }) {
        pair.MenuItem.Tag = pair.Value;
        pair.MenuItem.Click += delegate(object sender, EventArgs e) {
          SetTabulateMax((int)(sender as ToolStripMenuItem).Tag);
        };
      }

      // フォルダを選択して解析
      EventHandler visualizeSelectedDirectoryHandler = delegate {
        Visualize(SelectVisualizeDirectory());
      };

      toolStripMenuItemVisualize_Directory.Click += visualizeSelectedDirectoryHandler;
      toolStripButtonVisualizeDirectory.Click += visualizeSelectedDirectoryHandler;

      // 初期操作ボタン
      buttonStartUpVisualize_SelectDirectory.Click += visualizeSelectedDirectoryHandler;
      buttonStartUpVisualize_MyDocuments.Click += delegate {
        var directory = FileSystem.GetDirectory(Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments));
        Visualize(directory);
      };

      // 中断
      EventHandler abortVisualizeHandler = delegate {
        AbortVisualize();
      };

      toolStripMenuItemVisualize_Abort.Click += abortVisualizeHandler;
      toolStripButtonAbortVisualize.Click += abortVisualizeHandler;

      // 最新の状態に更新
      EventHandler refreshDirectoryTreeHandler = delegate {
        RefreshDirectoryTree();
      };

      toolStripMenuItemVisualize_Refresh.Click += refreshDirectoryTreeHandler;
      toolStripButtonRefresh.Click += refreshDirectoryTreeHandler;

      // SpecialFolderを解析
      var specialFolders = Runtime.IsRunningOnWindows
        ? new[] {
            Environment.SpecialFolder.Desktop,
            Environment.SpecialFolder.MyDocuments,
            Environment.SpecialFolder.ProgramFiles,
          }
        : new[] {
            Environment.SpecialFolder.Desktop,
            Environment.SpecialFolder.MyDocuments,
          };

      foreach (var specialFolder in specialFolders) {
        toolStripMenuItemVisualize.DropDownItems.Add(new DirectoryToolStripMenuItem(specialFolder, delegate(object sender, EventArgs e) {
          Visualize((sender as IDirectoryContainer).Directory);
        }));

        toolStripMain.Items.Add(new DirectoryToolStripMenuItem(specialFolder, delegate(object sender, EventArgs e) {
          Visualize((sender as IDirectoryContainer).Directory);
        }));
      }

      toolStripMenuItemVisualize.DropDownItems.Add(new ToolStripSeparator());

      toolStripMain.Items.Add(new ToolStripSeparator());

      // デバイスを解析
      foreach (var drive in FileSystem.GetDrives()) {
        if (drive.Type == System.IO.DriveType.CDRom || drive.Type == System.IO.DriveType.Removable)
          continue;

        toolStripMenuItemVisualize.DropDownItems.Add(new DriveToolStripMenuItem(drive, delegate(object sender, EventArgs e) {
          Visualize((sender as IDriveContainer).Drive);
        }));

        toolStripMain.Items.Add(new DriveToolStripButton(drive, delegate(object sender, EventArgs e) {
          Visualize((sender as IDriveContainer).Drive);
        }));
      }

      // 後ろに履歴を追加するためのセパレータ
      toolStripMain.Items.Add(new ToolStripSeparator());

      // 履歴のクリア
      toolStripMenuItemVisualize_History_Clear.Click += delegate { ClearVisualizedHistory(); };
 
      // 一つ上のフォルダへ
      toolStripButtonAscendDirectory.Click += delegate { AscendDirectory(); };

      // 戻る・進む
      toolStripSplitButtonBackHistory.ButtonClick += delegate { BackHistory(1); };
      toolStripSplitButtonForwardHistory.ButtonClick += delegate { ForwardHistory(1); };

      // About
      toolStripMenuItemHelp_About.Click += delegate {
        using (var dialog = new AboutDialog()) {
          dialog.ShowDialog(this);
        }
      };

      // 検索
      buttonSearchQuery.Click += delegate(object sender, EventArgs e) {
        contextMenuStripSearchQuery.Show(buttonSearchQuery,
                                         buttonSearchQuery.Width / 2,
                                         buttonSearchQuery.Height / 2);
      };

      foreach (var pair in new[] {
        new {Text = "画像ファイル", Pattern = @"\.(bmp|jpg|jpeg|png|gif)$"},
        new {Text = "アーカイブファイル", Pattern = @"\.(zip|lzh|cab|rar|gz|tgz|tbz|tar\.gz|tar\.bz2)$"},
        new {Text = "音声・動画ファイル", Pattern = @"\.(avi|flv|mov|mp3|mp4|mpg|mpeg|ogg|wav|wma|wmv)$"},
        new {Text = "ジャンクファイル", Pattern = @"(\.tmp|\.bak|\.old|\.log|~)$"},
      }) {
        var item = new SearchQueryRegexToolStripMenuItem(pair.Text, pair.Pattern);

        item.Click += delegate(object sender, EventArgs e) {
          var i = sender as SearchQueryRegexToolStripMenuItem;

          radioButtonSearchRegex.Checked = true;
          textBoxSearch.Text = i.Pattern;
          SearchByRegex(i.Pattern);
        };

        contextMenuStripSearchQuery.Items.Add(item);
      }

      toolStripMenuItemSearchQuery_Clear.Click += delegate {
        textBoxSearch.Text = string.Empty;
        ClearSearch();
      };

      textBoxSearch.Entered += delegate(object sender, EventArgs e) {
        var query = (sender as QueryTextBox).Text;

        if (string.IsNullOrEmpty(query))
          ClearSearch();
        else if (radioButtonSearchRegex.Checked)
          SearchByRegex(query);
        else if (radioButtonSearchWildCard.Checked)
          SearchByWildcard(query);
      };

      // Viewer関連
      viewer.ElementAtCursorChanged += delegate(object sender, ViewerMouseEventArgs e) {
        var v = sender as Viewer;

        if (e.Element == null) {
          toolTipElementAtCursor.SetToolTip(v.ViewingControl, null);

          SetStatistics(v.Graph.DirectoryTree);
        }
        else if (e.Element.IsFreeSpace) {
          SetStatistics(v.Graph.DirectoryTree.Drive);
        }
        else {
          var baseDirectoryName = v.Graph.DirectoryTree.FullName + System.IO.Path.DirectorySeparatorChar;
          var fullName = e.Element.Entry.FullName;

          if (e.Element.Entry is Directory)
            fullName += System.IO.Path.DirectorySeparatorChar;

          toolTipElementAtCursor.SetToolTip(v.ViewingControl,
                                            Smdn.IO.PathUtils.GetRelativePath(baseDirectoryName, fullName));

          SetStatistics(e.Element);
        }
      };


      viewer.ElementClicked += delegate(object sender, ViewerMouseEventArgs e) {
        if (e.Element.IsFreeSpace)
          return;

        var treeElement = e.Element as ITreeGraphElement;

        if ((e.Button & MouseButtons.Left) != MouseButtons.None) {
          // ディレクトリがクリックされた場合は、そのディレクトリを見る
          if (treeElement != null)
            ViewDirectory(treeElement);
        }
        else if ((e.Button & MouseButtons.Right) != MouseButtons.None) {
          var contextMenuStrip =(treeElement == null) ? fileContextMemuStrip : directoryContextMemuStrip;

          foreach (ToolStripItem item in contextMenuStrip.Items) {
            if (item is TabulatorToolStripMenuItem)
              item.Visible = false;
          }

          contextMenuStrip.Show((sender as Viewer).ViewingControl, e.X, e.Y, e.Element);
        }
      };
      viewer.MouseClicked += delegate(object sender, MouseEventArgs e) {
        if ((e.Button & MouseButtons.Right) != MouseButtons.None) {
          var v = sender as Viewer;

          if (v.Graph == null)
            return;

          rootDirectoryContextMemuStrip.Show(v.ViewingControl, e.X, e.Y, v.Graph.DirectoryTree);
        }
      };

      viewer.StatusChanged += delegate(object sender, ViewerStatusEventArgs e) { SetFinalStatus(e.Status); };
      viewer.HistoryChanged += delegate { HistoryChanged(); };
      viewer.GraphRootChanged += delegate { GraphRootChanged(); };

      // Tabulator関連
      tabulator.ItemClicked += delegate(object sender, TabulatorMouseEventArgs e) {
        if ((e.Button & MouseButtons.Left) != MouseButtons.None) {
          // do nothing
        }
        else if ((e.Button & MouseButtons.Right) != MouseButtons.None) {
          var contextMenuStrip = (e.Element is ITreeGraphElement) ? directoryContextMemuStrip : fileContextMemuStrip;

          foreach (ToolStripItem item in contextMenuStrip.Items) {
            if (item is TabulatorToolStripMenuItem)
              item.Visible = true;
          }

          contextMenuStrip.Show((sender as Tabulator).TabulatingListView, e.X, e.Y, e.Element);
        }
      };

      tabulator.SelectedElementChanged += delegate { SelectedElementChanged(); };

      // Viewer・Tabulatorを右クリックしたときに表示されるコンテキストメニュー
      fileContextMemuStrip = new GraphElementContextMenuStrip();
      directoryContextMemuStrip = new GraphElementContextMenuStrip();
      rootDirectoryContextMemuStrip = new DirectoryTreeContextMenuStrip();

      System.ComponentModel.CancelEventHandler contextMenuStripOpening = delegate(object sender, System.ComponentModel.CancelEventArgs e) {
        var de = sender as IDirectoryEntryContextMenuStrip;

        foreach (ToolStripItem item in (sender as ContextMenuStrip).Items) {
          if (item is ParentDirectoryEntryToolStripMenuItem)
            item.Available = (de.DirectoryEntry.Parent != null);

          var entryItem = item as DirectoryEntryToolStripMenuItem;

          if (entryItem != null)
            entryItem.UpdateText(de.DirectoryEntry);
        }
      };

      fileContextMemuStrip.Opening += contextMenuStripOpening;
      directoryContextMemuStrip.Opening += contextMenuStripOpening;
      rootDirectoryContextMemuStrip.Opening += contextMenuStripOpening;

      for (int i = 0; i < 3; i++) {
        ContextMenuStrip strip = null;
        ToolStripItem[] items = null;

        switch (i) {
          case 0:
            strip = fileContextMemuStrip;
            items = new ToolStripItem[] {
              new DirectoryEntryToolStripMenuItem("最新の状態に更新", delegate(object sender, EventArgs e) {
                RefreshDirectoryTree();
              }),
              new ToolStripSeparator(),
              new ParentDirectoryEntryToolStripMenuItem("親フォルダ \"{1}\" を見る", delegate(object sender, EventArgs e) {
                ViewDirectory(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Parent);
              }),
              new ToolStripSeparator(),
              new DirectoryEntryToolStripMenuItem("ファイル \"{0}\" を開く", delegate(object sender, EventArgs e) {
                Smdn.IO.FileManager.Open(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry.FullName);
              }),
              new ParentDirectoryEntryToolStripMenuItem("親フォルダ \"{1}\" を開く", delegate(object sender, EventArgs e) {
                Smdn.IO.FileManager.Browse(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry.FullName, true);
              }),
              new ToolStripSeparator(),
              new DirectoryEntryToolStripMenuItem("ファイル名をコピー", delegate(object sender, EventArgs e) {
                Clipboard.SetText(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry.Name);
              }),
              new DirectoryEntryToolStripMenuItem("フルパスをコピー", delegate(object sender, EventArgs e) {
                Clipboard.SetText(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry.FullName);
              }),
              new TabulatorToolStripMenuItem("すべてのファイル・フォルダのパスをコピー", delegate(object sender, EventArgs e) {
                CopyTabulatingElementPathsToClipboard();
              }),
              new ToolStripSeparator(),
              new DirectoryEntryToolStripMenuItem("ファイル \"{0}\" のプロパティ", delegate(object sender, EventArgs e) {
                Smdn.IO.FileManager.ShowProperty(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry.FullName);
              }),
            };
            break;

          case 1:
            strip = directoryContextMemuStrip;
            items = new ToolStripItem[] {
              new DirectoryEntryToolStripMenuItem("最新の状態に更新", delegate(object sender, EventArgs e) {
                //RefreshDirectoryTree((((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element as ITreeGraphElement).DirectoryTree);
                RefreshDirectoryTree();
              }),
              new ToolStripSeparator(),
              new DirectoryEntryToolStripMenuItem("フォルダ \"{0}\" を見る", delegate(object sender, EventArgs e) {
                ViewDirectory(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element as ITreeGraphElement);
              }),
              new ParentDirectoryEntryToolStripMenuItem("親フォルダ \"{1}\" を見る", delegate(object sender, EventArgs e) {
                ViewDirectory(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Parent);
              }),
              new ToolStripSeparator(),
              new DirectoryEntryToolStripMenuItem("フォルダ \"{0}\" を開く", delegate(object sender, EventArgs e) {
                Smdn.IO.FileManager.Browse(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry.FullName, false);
              }),
              new ParentDirectoryEntryToolStripMenuItem("親フォルダ \"{1}\" を開く", delegate(object sender, EventArgs e) {
                var directory = ((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry as Directory;

                Smdn.IO.FileManager.Browse(directory.FullName, (directory.Parent != null));
              }),
              new ToolStripSeparator(),
              new DirectoryEntryToolStripMenuItem("フォルダ名をコピー", delegate(object sender, EventArgs e) {
                Clipboard.SetText(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry.Name);
              }),
              new DirectoryEntryToolStripMenuItem("フルパスをコピー", delegate(object sender, EventArgs e) {
                Clipboard.SetText(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry.FullName);
              }),
              new TabulatorToolStripMenuItem("すべてのファイル・フォルダのパスをコピー", delegate(object sender, EventArgs e) {
                CopyTabulatingElementPathsToClipboard();
              }),
              new ToolStripSeparator(),
              new DirectoryEntryToolStripMenuItem("フォルダ \"{0}\" のプロパティ", delegate(object sender, EventArgs e) {
                Smdn.IO.FileManager.ShowProperty(((sender as ToolStripMenuItem).Owner as GraphElementContextMenuStrip).Element.Entry.FullName);
              }),
            };
            break;

          case 2:
            strip = rootDirectoryContextMemuStrip;
            items = new ToolStripItem[] {
              new DirectoryEntryToolStripMenuItem("最新の状態に更新", delegate(object sender, EventArgs e) {
                //RefreshDirectoryTree(((sender as ToolStripMenuItem).Owner as DirectoryTreeContextMenuStrip).DirectoryTree);
                RefreshDirectoryTree();
              }),
              new ToolStripSeparator(),
              new DirectoryEntryToolStripMenuItem("フォルダ \"{0}\" を開く", delegate(object sender, EventArgs e) {
                Smdn.IO.FileManager.Browse(((sender as ToolStripMenuItem).Owner as DirectoryTreeContextMenuStrip).DirectoryTree.FullName, false);
              }),
              new ParentDirectoryEntryToolStripMenuItem("親フォルダ \"{1}\" を開く", delegate(object sender, EventArgs e) {
                var directory = ((sender as ToolStripMenuItem).Owner as DirectoryTreeContextMenuStrip).DirectoryTree;

                Smdn.IO.FileManager.Browse(directory.FullName, (directory.Parent != null));
              }),
              new ToolStripSeparator(),
              new DirectoryEntryToolStripMenuItem("フォルダ名をコピー", delegate(object sender, EventArgs e) {
                Clipboard.SetText(((sender as ToolStripMenuItem).Owner as DirectoryTreeContextMenuStrip).DirectoryTree.Name);
              }),
              new DirectoryEntryToolStripMenuItem("フルパスをコピー", delegate(object sender, EventArgs e) {
                Clipboard.SetText(((sender as ToolStripMenuItem).Owner as DirectoryTreeContextMenuStrip).DirectoryTree.FullName);
              }),
              new ToolStripSeparator(),
              new DirectoryEntryToolStripMenuItem("フォルダ \"{0}\" のプロパティ", delegate(object sender, EventArgs e) {
                Smdn.IO.FileManager.ShowProperty(((sender as ToolStripMenuItem).Owner as DirectoryTreeContextMenuStrip).DirectoryTree.FullName);
              }),
            };
            break;
        }

        strip.Items.AddRange(items);
      }
    }
  }
}
