//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2003-2014 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Collections.Generic;

using Smdn.Collections;
using Smdn.IO.FileSystems;

namespace Smdn.Applications.BittenBaum {
  public class DirectoryTree : Directory {
    private class Fraction {
      public int Numerator;
      public readonly int Denominator;

      public Fraction(int numerator, int denominator)
      {
        this.Numerator = numerator;
        this.Denominator = denominator;
      }
    }

    /*
     * class members
     */
    public static DirectoryTree Create(Directory rootDirectory, Predicate<DirectoryEntry> match, Action<Directory, float> reportProgressCallback)
    {
      if (rootDirectory == null)
        throw new ArgumentNullException("rootDirectory");

      var tree = new DirectoryTree(rootDirectory, null);
      var progressFractionStack = new List<Fraction>();

      reportProgressCallback(tree.directory, 0.0f);

      tree.Scan(match);

      progressFractionStack.Add(new Fraction(0, tree.subTrees.Count));

      ScanCore(tree, match, progressFractionStack, reportProgressCallback);

      return tree;
    }

    private static void ScanCore(DirectoryTree tree, Predicate<DirectoryEntry> match, List<Fraction> progressFractionStack, Action<Directory, float> reportProgressCallback)
    {
      var progressPercentage = 0.0f;

      for (var index = progressFractionStack.Count - 1; 0 <= index; index--) {
        progressPercentage = (progressFractionStack[index].Numerator + progressPercentage) / (float)progressFractionStack[index].Denominator;
      }

      reportProgressCallback(tree.directory, progressPercentage);

      foreach (var subTree in tree.subTrees) {
        subTree.Scan(match);

        if (0 < subTree.subTrees.Count) {
          progressFractionStack.Add(new Fraction(0, subTree.subTrees.Count));

          ScanCore(subTree, match, progressFractionStack, reportProgressCallback);

          progressFractionStack.RemoveAt(progressFractionStack.Count - 1);
        }

        progressFractionStack[progressFractionStack.Count - 1].Numerator++;
      }
    }

    private static void ScanDirectoryExceptionHandler(Directory directory, Exception exception)
    {
      if (exception is UnauthorizedAccessException)
        Console.Error.WriteLine("UnauthorizedAccessException in {0}", directory.FullName);
      else if (exception is System.IO.DirectoryNotFoundException)
        Console.Error.WriteLine("DirectoryNotFoundException in {0}", directory.FullName);
      else if (exception is System.IO.IOException)
        Console.Error.WriteLine("IOException in {0}", directory.FullName);
      else
        throw exception;
    }

    /*
     * instance members
     */
    public Directory Directory {
      get { return directory; }
    }

    public DirectoryTree ParentTree {
      get { return parentTree; }
    }

    public int Depth {
      get { return (parentTree == null) ? 0 : parentTree.Depth + 1; }
    }

    public int MaxDepth {
      get
      {
        var maxDepth = 1;

        foreach (var subtree in subTrees) {
          if (maxDepth <= subtree.MaxDepth)
            maxDepth = subtree.MaxDepth + 1;
        }

        return maxDepth;
      }
    }

    private readonly Directory directory;
    private readonly DirectoryTree parentTree;
    private List<DirectoryTree> subTrees = new List<DirectoryTree>();
    private List<File> files = new List<File>();

    private DirectoryTree(Directory directory, DirectoryTree parentTree)
    {
      this.directory = directory;
      this.parentTree = parentTree;
    }

    private void Scan(Predicate<DirectoryEntry> match)
    {
#if DEBUG
      if (0 < subTrees.Count || 0 < files.Count)
        throw new InvalidOperationException("already scanned");
#endif

      foreach (var entry in directory.GetEntries(ScanDirectoryExceptionHandler, match)) {
        var file = entry as File;

        if (file == null)
          subTrees.Add(new DirectoryTree(entry as Directory, this));
        else
          files.Add(file);
      }
    }

#region "members from Smdn.IO.FileSystem.Directory"
    public override string Name {
      get { return directory.Name; }
    }

    public override long Length {
      get { return directory.Length; }
    }

    public override bool Exists {
      get { return directory.Exists; }
    }

    public override Directory Parent {
      get { return directory.Parent; }
    }

    public override Directory Root {
      get { return directory.Root; }
    }

    public override bool IsRoot {
      get { return directory.IsRoot; }
    }

    public override Drive Drive {
      get { return directory.Drive; }
    }

    public override string FullName {
      get { return directory.FullName; }
    }

    public override object RawInfo {
      get { return directory.RawInfo; }
    }

    public override IEnumerable<Directory> GetDirectories(ScanDirectoryExceptionHandler exceptionHandler, Predicate<Directory> match)
    {
      return directory.GetDirectories(exceptionHandler, match);
    }

    public override IEnumerable<DirectoryEntry> GetEntries(ScanDirectoryExceptionHandler exceptionHandler, Predicate<DirectoryEntry> match)
    {
      return directory.GetEntries(exceptionHandler, match);
    }

    public override IEnumerable<File> GetFiles(ScanDirectoryExceptionHandler exceptionHandler, Predicate<File> match)
    {
      return directory.GetFiles(exceptionHandler, match);
    }

    public override long GetLength()
    {
      return directory.GetLength();
    }
#endregion

    /*
     * entry count properties
     */
    public int EntryCount {
      get { return files.Count + subTrees.Count; }
    }

    private int? totalEntryCount = null;

    public int TotalEntryCount {
      get
      {
        if (totalEntryCount.HasValue)
          return totalEntryCount.Value;

        totalEntryCount = EntryCount;

        foreach (var subtree in subTrees) {
          totalEntryCount += subtree.TotalEntryCount;
        }

        return totalEntryCount.Value;
      }
    }

    private int? totalFileCount = null;

    public int TotalFileCount {
      get
      {
        if (totalFileCount.HasValue)
          return totalFileCount.Value;

        totalFileCount = files.Count;

        foreach (var subtree in subTrees) {
          totalFileCount += subtree.TotalFileCount;
        }

        return totalFileCount.Value;
      }
    }

    private int? totalSubTreeCount = null;

    public int TotalSubTreeCount {
      get
      {
        if (totalSubTreeCount.HasValue)
          return totalSubTreeCount.Value;

        totalSubTreeCount = 0;

        foreach (var subtree in subTrees) {
          totalSubTreeCount++; // subtree itself
          totalSubTreeCount += subtree.TotalSubTreeCount;
        }

        return totalSubTreeCount.Value;
      }
    }

    private long? totalFileSize;

    public long GetTotalFileSize()
    {
      if (totalFileSize.HasValue)
        return totalFileSize.Value;

      var a = 0L;

      foreach (var subTree in subTrees) {
        a += subTree.GetTotalFileSize();
      }

      foreach (var file in files) {
        a += file.Length;
      }

      totalFileSize = a;

      return a;
    }

    public void Refresh(Predicate<DirectoryEntry> match, Action<Directory, float> reportProgressCallback)
    {
      totalEntryCount = null;
      totalFileCount = null;
      totalSubTreeCount = null;
      totalFileSize = null;

      var refreshedTree = Create(directory, match, reportProgressCallback);

      this.subTrees = refreshedTree.subTrees;
      this.files = refreshedTree.files;
    }

    public IEnumerable<DirectoryEntry> GetSortedEnumerator(Comparison<DirectoryEntry> comparison)
    {
      if (comparison == null)
        throw new ArgumentNullException("comparison");

      var entries = new List<DirectoryEntry>(files.Count + subTrees.Count);

      //entries.AddRange(files);
      foreach (var file in files) {
        entries.Add(file);
      }

      //entries.AddRange(subTrees);
      foreach (var subTree in subTrees) {
        entries.Add(subTree);
      }

      entries.Sort(comparison);

      return entries;
    }

#if DEBUG
    public void Traverse(Action<DirectoryTree> action)
    {
      action(this);

      foreach (var subtree in subTrees) {
        subtree.Traverse(action);
      }
    }

    public void Traverse(Action<DirectoryEntry> action)
    {
      action(directory);

      foreach (var entry in files) {
        action(entry);
      }

      foreach (var subtree in subTrees) {
        subtree.Traverse(action);
      }
    }
#endif
  }
}
